from collections import namedtuple
from itertools import repeat
import warnings
import logging
import os.path as op
import numpy as np
from mlbspline import load
from lbftd import evalGibbs as eg
from lbftd.statevars import iP, iT

log = logging.getLogger('seafreeze')
stream = logging.StreamHandler()
stream.setFormatter(logging.Formatter('[SeaFreeze %(levelname)s] %(message)s'))
defpath = op.join(op.dirname(op.abspath(__file__)), 'SeaFreeze_Gibbs.mat')

def seafreeze(PT, phase, path=defpath):
    """ Calculates thermodynamic quantities for H2O water or ice polymorphs Ih, III, V, and VI for all phases
        (see lbftd documentation for full list)
        for solid phases only:
            - Vp (compressional wave velocity, in m/s)
            - Vs (shear wave velocity, in m/s)
            - shear (shear modulus, in MPa)
    Requires the SeaFreeze_Gibbs.mat library containing the Gibbs LBF parametrization (installed with this module).

    NOTE:  The authors recommend the use of 'water1' for any application in the 200-355 K range and up to 2300 MPa.
    The ice Gibbs parametrizations are optimized for use with the 'water1' phase for phase equilibrium calculations.
    Using other water parametrizations will lead to incorrect melting curves -- 'water2' and 'water_IAPWS95'
    parametrizations are provided for HP extension up to 100 GPa and comparison only.

    :param PT:      The pressure (MPa) and temperature (K) conditions at which the thermodynamic quantities should be
                    calculated -- the specified units are required, as conversions are built into several calculations.
                    This parameter can have one of the following formats:
                        - A 1-dimensional numpy array of tuples with one or more scattered (P,T) tuples, e.g.
                                PT = np.empty((3,), np.object)
                                PT[0] = (441.0858, 313.95)
                                PT[1] = (478.7415, 313.96)
                                PT[2] = (444.8285, 313.78)
                        - A numpy array with 2 nested numpy arrays, the first with pressures and the second
                          with temperatures -- each inner array must be sorted from low to high values
                          A grid will be constructed from the P and T arrays such that each row of the output
                          will correspond to a pressure and each column to a temperature, e.g.
                                P = np.arange(0.1, 1000.2, 10)
                                T = np.arange(240, 501, 2)
                                PT = np.array([P, T])
    :param phase:   One of the keys of the phases dict, indicating the phase of H2O to be evaluated
    :param path:    An optional path to the SeaFreeze_Gibbs.mat file
                    default value assumes the spline distributed along with the project
    :return:        object containing the calculated thermodynamic quantities (as named properties), as well as
                    A PTM property (a copy of PT)
    """
    try:
        phasedesc = phases[phase]
    except KeyError:
        raise ValueError('The specified phase is not recognized. Supported phases are ' +
                         ', '.join(phases.keys())+'.')
    sp = load.loadSpline(path, phasedesc.sp_name)
    # Calc density and isentropic bulk modulus
    isscatter = _is_scatter(PT)
    tdvs = _get_tdvs(sp, PT, isscatter)
    if phasedesc.shear_mod_parms:
        smg = _get_shear_mod_GPa(phasedesc.shear_mod_parms, tdvs.rho, _get_T(PT, isscatter))
        tdvs.shear = 1e3 * smg  # Convert to MPa for consistency with other measures
        tdvs.Vp = _get_Vp(smg, tdvs.rho, tdvs.Ks)
        tdvs.Vs = _get_Vs(smg, tdvs.rho)
    return tdvs


def whichphase(PT, path=defpath):
    """ Determines the most likely phase of water at each pressure/temperature

    :param PT:      The pressure (MPa) and temperature (K) conditions at which the phase should be determined --
                    the specified units are required, as conversions are built into several calculations.
                    This parameter can have one of the following formats:
                        - a 1-dimensional numpy array of tuples with one or more scattered (P,T) tuples, e.g.
                                PT = np.empty((3,), np.object)
                                PT[0] = (441.0858, 313.95)
                                PT[1] = (478.7415, 313.96)
                                PT[2] = (444.8285, 313.78)
                        - a numpy array with 2 nested numpy arrays, the first with pressures and the second
                          with temperatures -- each inner array must be sorted from low to high values
                          a grid will be constructed from the P and T arrays such that each row of the output
                          will correspond to a pressure and each column to a temperature, e.g.
                                P = np.arange(0.1, 1000.2, 10)
                                T = np.arange(240, 501, 2)
                                PT = np.array([P, T])
    :param path:    An optional path to the SeaFreeze_Gibbs.mat file --
                    default value assumes the spline distributed along with the project
    :return:        A numpy.ndarray the same size as PT, with the phase of each pressure/temperature represented by
                    an integer, as shown in phasenum2phase
    """
    isscatter = _is_scatter(PT)
    phase_sp = {v.phase_num: load.loadSpline(path, v.sp_name) for v in phases.values() if not np.isnan(v.phase_num)}
    ptsh = ((PT.size,) if isscatter else (PT[0].size, PT[1].size))  # Reference shape based on PT
    comp = np.full(ptsh + (max_phase_num+1,), np.nan)  # Comparison matrix
    for p in phase_sp.keys():
        sl = tuple(repeat(slice(None), 1 if isscatter else 2))+(p,)  # Slice for this phase
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            sp = phase_sp[p]
            tdvs = _get_tdvs(sp, PT, isscatter, 'G').G
            # Wipe out G for PT values that fall outside the knot sequence
            if isscatter:
                extrap = [(pt[iP] < sp['knots'][iP].min()) + (pt[iP] > sp['knots'][iP].max()) +
                          (pt[iT] < sp['knots'][iT].min()) + (pt[iT] > sp['knots'][iT].max()) for pt in PT]
            else:
                pt = np.logical_or(PT[iP] < sp['knots'][iP].min(), PT[iP] > sp['knots'][iP].max())
                tt = np.logical_or(PT[iT] < sp['knots'][iT].min(), PT[iT] > sp['knots'][iT].max())
                extrap = np.logical_or(*np.meshgrid(pt, tt, indexing='ij'))
            tdvs[extrap] = np.nan
            comp[sl] = tdvs
    # Output for all-nan slices should be nan
    all_nan_sl = np.all(np.isnan(comp), -1)  # Find slices where all values are nan along the innermost axis
    out = np.full(ptsh, np.nan)  # Initialize output to nan
    out[~all_nan_sl] = np.nanargmin(comp[~all_nan_sl],-1)  # Find min values for other slices
    return out


def _get_tdvs(sp, PT, is_scatter, *tdvSpec):
    """ Peeks into PT to see if the PT data is for grid or scatter and calls the appropriate evalGibbs function

    :param sp:          The Gibbs LBF
    :param PT:          The PT data
    :param is_scatter:  Boolean indicating whether the PT data is scatter or not (if not, it is a grid)
    :param tdvSpec:     optional list of thermodynamic variables to calculate (see lbftd documentation)
    :return:            tdv object (see lbftd documentation)
    """
    fn = eg.evalSolutionGibbsScatter if is_scatter else eg.evalSolutionGibbsGrid
    return fn(sp, PT, *tdvSpec, failOnExtrapolate=False)


def _get_shear_mod_GPa(sm, rho, T):
    return None if sm is None else sm[0] + sm[1]*(rho - sm[4]) + sm[2]*(rho-sm[4])**2 + sm[3]*(T-sm[5])


def _get_Vp(smg, rho, Ks):
    return 1e3 * np.sqrt((Ks/1e3 + 4/3*smg)/rho/1e-3)


def _get_Vs(smg, rho):
    return 1e3 * np.sqrt(smg/rho/1e-3)


def _is_scatter(PT):
    return isinstance(PT[0], tuple) or (PT.shape == (1,2) and np.isscalar(PT[0]) and np.isscalar(PT[1]))


def _get_T(PT, is_scatter):
    return np.array([T for P,T in PT]) if is_scatter else PT[1]


#########################################
## Constants
#########################################
PhaseDesc = namedtuple('PhaseDesc', 'sp_name shear_mod_parms phase_num')
phases = {"Ih": PhaseDesc("G_iceIh", [3.1, -0.00462, 0, -0.00657, 1000, 273.15], 1),  # Feistel and Wagner, 2006
          "II": PhaseDesc("G_iceII", [4.1, 0.0175, 0, -0.014, 1100, 273], 2),          # Journaux et al, 2019
          "III": PhaseDesc("G_iceIII", [2.57, 0.0175, 0, -0.014, 1100, 273], 3),       # Journaux et al, 2019
          "V": PhaseDesc("G_iceV", [2.57, 0.0175, 0, -0.014, 1100, 273], 5),           # Journaux et al, 2019
          "VI": PhaseDesc("G_iceVI", [2.57, 0.0175, 0, -0.014, 1100, 273], 6),         # Journaux et al, 2019
          "water1": PhaseDesc("G_H2O_2GPa_500K", None, 0),              # extends to 500 K and 2300 MPa; Bollengier et al 2019
          "water2": PhaseDesc("G_H2O_100GPa_10000K", None, np.nan),     # extends to 100 GPa; Brown 2018
          "water_IAPWS95": PhaseDesc("G_H2O_IAPWS", None, np.nan)       # LBF representation of IAPWS 95; Wagner and Pruß, 2002
          }
max_phase_num = max([p.phase_num for p in phases.values()])
phasenum2phase = {v.phase_num: k for (k,v) in phases.items() if not np.isnan(v.phase_num)}



from scipy.io import loadmat
import numpy as np
import logging

log = logging.getLogger('mlbspline')
stream = logging.StreamHandler()
stream.setFormatter(logging.Formatter('[MLBspline %(levelname)s] %(message)s'))

def loadSpline(splineFile, splineVar=None):
    """Loads a spline from .mat format file
    Only MatLab files saved as either v7 or v7.3 are supported

    :param splineFile: full or relative path to Matlab file
    :param splineVar: variable to load from splineFile
    :return: a dict with the spline representation
    """
    raw = _getRaw(splineFile, splineVar)
    spd = getSplineDict(_stripNestingToFields(raw))
    # Knots can sometimes be imported as object type, which cannot be automatically cast to float
    if not np.issubdtype(spd['knots'][0].dtype, np.number) and not isinstance(spd['knots'][0][0], np.ndarray):
        spd['knots'] = spd['knots'].astype(np.float_)
    validateSpline(spd)
    return spd


def _getRaw(splineFile, splineVar=None):
    f = _loadFile(splineFile, splineVar)
    splineVar = _getCheckVar(f, splineVar)
    return f[splineVar]


def _loadFile(f, v=None):
    return loadmat(f, variable_names=None if v is None else [v], chars_as_strings=True)


def _getCheckVar(rawf, v=None):
    contents = [k for k in rawf.keys() if not k.startswith('__')]
    if v is None:
        if len(contents) == 1:
            v = contents[0]
        else:
            raise ValueError('The splineFile contains multiple variables: ' + ', '.join(contents) +
                             'Please provide the appropriate splineVar.')
    elif v not in contents:
        raise ValueError('The specified splineVar cannot be found in the specified file.  It only contains ' +
                         'variables ' + ', '.join(contents))
    return v


def _stripNestingToFields(src):
    lastFields = src.dtype.names
    while src[0].dtype.names == lastFields:
        src = src[0]
        lastFields = src.dtype.names
    return src


def _stripNestingToValue(src):
    # Numpy already automatically squeezes
    while isinstance(src, np.ndarray) and src.shape[0] == 1:
        src = src[0]
    return src


def getSplineDict(src):
    # All splines must contain the following fields
    out = {
        'form':     _stripNestingToValue(src['form']),
        'knots':    np.array([_stripNestingToValue(kd) for kd in _stripNestingToValue(src['knots'])], dtype=object),
        'number':   _stripNestingToValue(src['number']).astype(int),
        'order':    _stripNestingToValue(src['order']).astype(int),
        'dim':      _stripNestingToValue(src['dim']).astype(int),
        'coefs':    _stripNestingToValue(src['coefs']),
        'ndT':      False
    }

    # If dimensionless temperature is used, include the critical temperature in field ['Tc'] in the outputs to flag it
    if 'Tc' in src.dtype.names:
        out['Tc'] = _stripNestingToValue(src['Tc'])
        out['ndT'] = True

    # If number is a scalar, this is a 1D spline and some stuff needs to be re-wrapped for later code to work
    if not isinstance(out['number'], np.ndarray):
        out['number'] = np.array([out['number']])
        out['order'] = np.array([out['order']])
        # We can't just reshape or nest in one step, sadly. Numpy tries to be smart and creates a 2D array.
        # See https://stackoverflow.com/questions/7667799/numpy-object-arrays
        knots = np.empty(1, object)
        knots[0] = out['knots']
        out['knots'] = knots

    return out


def validateSpline(spd):
    """Checks the spline representation to make sure it has all the expected fields for evaluation
    Throws an error if anything is missing (one at a time)

    :param spd: a dict (output from getSplineDict) representing the spline
    """
    # Currently only supports b-splines
    if spd['form'] != 'B-':
        raise ValueError('These functions currently only support b-splines.')
    # Currently only supports 1-D output
    if spd['dim'] != 1:
        raise ValueError('These functions currently only support 1-D values.')
    # Need to have same size for knots, number, order
    if spd['number'].shape[0] != spd['knots'].shape[0] or spd['number'].shape != spd['order'].shape:
        raise ValueError('The spline''s knots, number, and order are inconsistent.')
    # Number of dims in coefs should be same as size of number (also makes it match knots and order b/c of prev check)
    if spd['number'].size != spd['coefs'].ndim:
        raise ValueError('The spline''s coefficients are inconsistent with its number.')
    # Each dim of coefs should have as many members as indicated by number
    if not (np.array(spd['coefs'].shape) == spd['number']).all():
        raise ValueError('At least one of the spline''s coefficients doesn''t match the corresponding number.')
    # Each dim of knots should have be of length dictated by corresponding number + order
    if not np.array(spd['number'] + spd['order'] == [k.size for k in spd['knots']]).all():
        raise ValueError('The number of knots does not correspond to the number and order for at least one variable.')
    return



from collections.__init__ import namedtuple
from inspect import getargvalues, currentframe
from pprint import pformat
import numpy as np
from numpy.lib.scimath import sqrt
import logging

from mlbspline.eval import evalMultivarSpline

log = logging.getLogger('lbftd')
stream = logging.StreamHandler()
stream.setFormatter(logging.Formatter('[LBFTD %(levelname)s] %(message)s'))

'''
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
Warning: units must be as specified in the README because some conversions are hardcoded.
With the exception of pressure, units are SI.  Pressure is in MPa rather than Pa.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
'''


#########################################
## TDV eval functions
#########################################

def evalGibbsEnergy(gibbsSp, PTM):
    """
    :return: G
    """
    return evalMultivarSpline(gibbsSp, PTM)


def evalIsobaricSpecificHeat(gibbsSp, gPTM, derivs):
    """
    :return: Cp
    """
    if gibbsSp['ndT']:
        T = np.exp(gPTM[iT])*gibbsSp['Tc']
    else:
        T = gPTM[iT]
    return -1 * derivs.d2T * T


def evalIsochoricSpecificHeat(tdv, gPTM, derivs):
    """

    :return: Cv
    """
    #Cv= Cp + Tm.*dPT.^2./d2P;
    return tdv.Cp + gPTM[iT] * np.power(derivs.dPT, 2) / derivs.d2P


def evalEntropy(derivs):
    """
    :return: S
    """
    return -1 * derivs.d1T


def evalSoundSpeed(derivs):
    """
    :return: vel
    """
    # MPa-Pa unit conversion cancels
    return np.real(sqrt(np.power(derivs.d1P, 2) / (np.power(derivs.dPT, 2) / derivs.d2T - derivs.d2P)))


def evalDensity(derivs):
    """
    :return: rho
    """
    return 1e6 * np.power(derivs.d1P, -1)  # 1e6 for MPa to Pa


def evalIsothermalBulkModulus(derivs):
    """
    :return: Kt
    """
    return -1 * derivs.d1P / derivs.d2P


def evalPDerivIsothermalBulkModulus(derivs):
    """
    :return: Kp (K')
    """
    return derivs.d1P * np.power(derivs.d2P, -2) * derivs.d3P - 1


def evalIsentropicBulkModulus(tdv):
    """
    :return: Ks
    """
    return tdv.rho * np.power(tdv.vel, 2) / 1e6 #  1e6 for MPa conversion


def evalThermalExpansivity(derivs, tdv):
    """
    :return: alpha
    """
    return 1e-6 * derivs.dPT * tdv.rho  #  1e-6 for MPa conversion


def evalInternalEnergy(gPTM, tdv):
    """
    :param gPTM: gridded dimensions over which spline is being evaluated
    :return:     U
    """
    return tdv.G - 1e6 * gPTM[iP] / tdv.rho + gPTM[iT] * tdv.S  #  1e6 for MPa conversion


def evalSoluteChemicalPotential(MWu, f, derivs, tdv):
    """
    :return: mus
    """
    return MWu * tdv.G + f * derivs.d1M


def evalSolventChemicalPotential(MWv, f, gPTM, derivs, tdv):
    """
    :return: muw
    """
    return (tdv.G * MWv) - (MWv * f * gPTM[iM] * derivs.d1M)


def evalHelmholtzEnergy(gPTM, tdv):
    """
    :return: A
    """
    return tdv.U - gPTM[iT] * tdv.S


def evalEnthalpy(gPTM, tdv):
    """
    :return: H
    """
    return tdv.G + gPTM[iT] * tdv.S


def evalPartialMolarVolume(MWu, f, derivs):
    """ Slope at a point of the V vs. M graph (volume vs molality)
    :return: Vm
    """
    return (MWu * derivs.d1P) + (f * derivs.dPM)


def evalPartialMolarHeatCapacity(MWu, f, tdv, derivs, gPTM):
    """
    :return: Cpm
    """
    return MWu * tdv.Cp - f * derivs.d2T1M * gPTM[iT]


def evalVolume(tdv):
    """
    :return: V
    """
    return np.power(tdv.rho, -1)


def evalApparentSpecificHeat(f, gPTM, tdv):
    """
    :return: Cpa
    """
    return (tdv.Cp * f - _get0MTdv(tdv, 'Cp')) / _getDividableBy(gPTM[iM])


def evalApparentVolume(f, gPTM, tdv):
    """ Slope of a chord between pure solvent and a concentration on a V v. M graph
    :return: Va
    """
    return 1e6 * (tdv.V * f - _get0MTdv(tdv, 'V')) / _getDividableBy(gPTM[iM]) # 1e6 for MPa conversion


def _get0MTdv(tdv, prop):
    slc = [slice(None)] * (len(tdv.PTM))
    slc[iM] = slice(0, 1)
    return getattr(tdv, prop)[tuple(slc)]


def _getDividableBy(inp):
    eps = np.finfo(inp.dtype).eps
    return np.where(inp != 0, inp, eps)


#########################################
## Set up derivatives
#########################################

def _getSupportedDerivatives():
    """ Define a name for a derivative of Gibbs energy
    and the derivative order with respect to P, T, and M required to calculate it
    (if not provided, all default to defDer)
    These can then be used in a TDVSpec

    :return: An immutable iterable containing a list of supported derivatives
    """
    return tuple([
        derivSpec('d1P', wrtP=1, wrtT=defDer, wrtM=defDer),
        derivSpec('d1T', wrtP=defDer, wrtT=1, wrtM=defDer),
        derivSpec('d1M', wrtP=defDer, wrtT=defDer, wrtM=1),
        derivSpec('dPT', wrtP=1, wrtT=1, wrtM=defDer),
        derivSpec('dPM', wrtP=1, wrtT=defDer, wrtM=1),
        derivSpec('d2P', wrtP=2, wrtT=defDer, wrtM=defDer),
        derivSpec('d2T', wrtP=defDer, wrtT=2, wrtM=defDer),
        derivSpec('d2T1M', wrtP=defDer, wrtT=2, wrtM=1),
        derivSpec('d3P', wrtP=3,wrtT=defDer, wrtM=defDer)
    ])


#########################################
## Set up thermodynamic variables (TDVs)
#########################################

def _getTDVSpec(name, calcFn, reqM=False, reqMWv=False, parmMWv='MWv', reqMWu=False, parmMWu='MWu',
                reqGrid=False, parmgrid='gPTM', reqF=False, parmf='f',
                reqDerivs=None, parmderivs='derivs', reqTDV=None, parmtdv='tdv', reqSpline=False,
                parmspline='gibbsSp', reqPTM=False, parmptm='PTM', req0M=False):
    """ Builds a TDVSpec namedtuple indicating what is required to calculate this particular thermodynamic variable
    :param name:        the name / symbol of the tdv (e.g., G, rho, alpha, muw)
    :param calcFn:      the name of the function used to calculate the tdv
    :param reqM:        True if DIRECTLY calculating the tdv requires concentration.  False otherwise
    :param reqMWv:      True if DIRECTLY calculating the tdv requires solvent molecular weight.  False otherwise
    :param parmMWv:     the name of the parameter of calcFn used to pass in solvent molecular weight if reqMWv
    :param reqMWu:      True if DIRECTLY calculating the tdv requires solute molecular weight.  False otherwise
    :param parmMWu:     the name of the parameter of calcFn used to pass in solute molecular weight if reqMWu
    :param reqGrid:     True if DIRECTLY calculating the tdv requires you to grid the PT[M] values.  False otherwise
    :param parmgrid:    the name of the parameter of calcFn used to pass in the gridded input dimensions if reqGrid
    :param reqF:        True if DIRECTLY calculating the tdv requires a conversion factor that gives the volume
                        of solvent in a unit volume of solution (this is calculated by
                        evalGibbs._getVolSolventInVolSlnConversion)
    :param parmf:       the name of the parameter of calcFn used to pass in the vol solvent to vol solution conversion
                        factor
    :param reqDerivs:   A list of derivatives needed to DIRECTLY calculate the tdv
                        (e.g. for tdv Kp, this would be ['d1P','dP','d3P']
                        - see fn evalPDerivIsothermalBulkModulus)
                        See getDerivatives for a full list of derivatives that can be calculated
    :param parmderivs:  the name of the parameter of calcFn used to pass in the pre-calculated derivatives if reqDerivs
    :param reqTDV:      A list of other thermodynamic variables needed to DIRECTLY calculate the tdv
                        (e.g. for tdv U, this would be ['G','rho'] - see fn evalInternalEnergy)
                        Note: recursive dependencies are not supported
    :param parmtdv:     the name of the parameter of calcFn used to pass in the tdvout if reqTDV
    :param reqSpline:   If True, calcFn needs the spline definition
    :param parmspline:  the name of the parameter of calcFn used to pass in the spline def if reqSpline
    :param reqPTM:      If True, calcFn needs the original dimension input (parm PTM in evalSolutionGibbs*) to run
    :param parmptm:     the name of the parameter of calcFn used to pass in the original input if reqPTM
    :param req0M:       if True, calcFn needs the 0 concentration for calculating apparent values
    :return:            a namedtuple giving the spec for the tdv
    """
    if reqDerivs is None:
        reqDerivs = []
    if reqTDV is None:
        reqTDV = []

    reqTDV = set(reqTDV); reqDerivs = set(reqDerivs)  # Make these sets to enforce uniqueness and immutability
    if name in reqTDV:
        raise ValueError('Recursive dependencies are not supported.  Amend the ' + name + ' TDVSpec accordingly')
    if reqDerivs.difference(derivativenames):
        raise ValueError('One or more derivatives are not supported. Amend the ' + name + ' TDVSpec accordingly.' +
                         'Supported derivatives are '+pformat(derivativenames))
    arginfo = getargvalues(currentframe())
    # Build properties of the TDVSpec dynamically
    flds = arginfo.args
    if not reqMWv:      flds.remove('parmMWv')
    if not reqMWu:      flds.remove('parmMWu')
    if not reqGrid:     flds.remove('parmgrid')
    if not reqDerivs:   flds.remove('parmderivs')
    if not reqTDV:      flds.remove('parmtdv')
    if not reqSpline:   flds.remove('parmspline')
    if not reqPTM:      flds.remove('parmptm')
    vals = {f: v for f, v in arginfo.locals.items() if f in flds}

    tdvspec = namedtuple('TDVSpec', flds)
    return tdvspec(**vals)


def _getSupportedThermodynamicVariables():
    """ When adding new tdvs, you don't need to worry about adding them in any particular order -
        dependencies are handled elsewhere
        See also the comments for _getTDVSpec and evalSolutionGibbs*

    :return: immutable iterable with the the full set of specs for supported thermodynamic variables
    """
    out = tuple([
        _getTDVSpec('G', evalGibbsEnergy, reqSpline=True, reqPTM=True),
        _getTDVSpec('U', evalInternalEnergy, reqGrid=True, reqTDV=['G', 'rho', 'S']),
        _getTDVSpec('A', evalHelmholtzEnergy, reqGrid=True, reqTDV=['U', 'S']),
        _getTDVSpec('H', evalEnthalpy, reqGrid=True, reqTDV=['G', 'S']),
        _getTDVSpec('S', evalEntropy, reqDerivs=['d1T']),
        _getTDVSpec('rho', evalDensity, reqDerivs=['d1P']),
        _getTDVSpec('V', evalVolume, reqTDV=['rho']),
        _getTDVSpec('Cp', evalIsobaricSpecificHeat, reqSpline=True, reqGrid=True, reqDerivs=['d2T']),
        _getTDVSpec('Cv', evalIsochoricSpecificHeat, reqGrid=True, reqDerivs=['dPT', 'd2P'], reqTDV=['Cp']),
        _getTDVSpec('Kt', evalIsothermalBulkModulus, reqDerivs=['d1P', 'd2P']),
        _getTDVSpec('Kp', evalPDerivIsothermalBulkModulus, reqDerivs=['d1P', 'd2P', 'd3P']),
        _getTDVSpec('Ks', evalIsentropicBulkModulus, reqTDV=['rho', 'vel']),
        _getTDVSpec('vel', evalSoundSpeed, reqDerivs=['d1P', 'dPT', 'd2T', 'd2P']),
        _getTDVSpec('alpha', evalThermalExpansivity, reqDerivs=['dPT'], reqTDV=['rho']),
        _getTDVSpec('mus', evalSoluteChemicalPotential, reqMWu=True, reqF=True, reqDerivs=['d1M'], reqTDV=['G']),
        _getTDVSpec('muw', evalSolventChemicalPotential, reqM=True, reqMWv=True, reqGrid=True, reqF=True,
                    reqDerivs=['d1M'], reqTDV=['G']),
        _getTDVSpec('Vm', evalPartialMolarVolume, reqMWu=True, reqF=True, reqDerivs=['d1P', 'dPM']),
        _getTDVSpec('Cpm', evalPartialMolarHeatCapacity, reqMWu=True, reqGrid=True, reqF=True, reqDerivs=['d2T1M'],
                    reqTDV=['Cp']),
        _getTDVSpec('Cpa', evalApparentSpecificHeat, reqM=True, reqGrid=True, reqF=True, reqTDV=['Cp'], req0M=True),
        _getTDVSpec('Va', evalApparentVolume, reqM=True, reqGrid=True, reqF=True, reqTDV=['V'], req0M=True)
            ])
    # Check that all reqTDVs are represented in the list
    outnames = frozenset([t.name for t in out])
    unrecognizedDependencies = [(tdv.name, dep) for tdv in out for dep in tdv.reqTDV if dep not in outnames]
    if unrecognizedDependencies:
        raise ValueError('One or more statevars depend on an unrecognized TDV: '+pformat(unrecognizedDependencies))
    return out, outnames


def _getPTOnlyTDVSpec():
    """ Determines the limited TDVs that can be calculated for a 2-D (P and T only) spline
    """
    # Concentration required if the flag says it does, if it uses F (which req concentration),
    # or if it uses a concentration derivative
    Mreq = [m for m in statevars if m.reqM or m.reqF or [d for d in m.reqDerivs if 'M' in d]]
    # Return values should be immutable - use tuples as TDVSpec namedtuples are not hashable so set is not usable
    return tuple([m for m in statevars if m not in Mreq])


#########################################
## Interpret and expand a spec
#########################################

def _addTDVDependencies(origTDVs):
    """ Recursively adds thermodynamic variables on which origTDVs are dependent
    This does not handle derivative dependencies - see getDerivatives

    :param origTDVs:    An iterable of TDVs to check for dependencies
                        elements can be either the string names of statevars or the TDV objects
    :return:            Immutable iterable of TDV objects including origTDVs and any TDVs on which they are
                        directly or indirectly dependent
    """
    otdvnames = set(origTDVs) if isinstance(next(iter(origTDVs)), str) else {m.name for m in origTDVs}
    while True:
        # Get TDVs dependent on TDVs in otdvnames
        reqot = {t for m in statevars if m.name in otdvnames for t in m.reqTDV}
        # If every tdv in reqot is already in otdvnames, stop
        if reqot.issubset(otdvnames):
            break
        else:
            # Add the new list of tdvs and repeat the loop to add more nested dependencies
            otdvnames = otdvnames.union(reqot)
    return tuple([m for m in statevars if m.name in otdvnames])


def expandTDVSpec(tdvSpec, dimCt):
    """ Add dependencies for requested thermodynamic variables or sets default tdvSpec if none provided
    derivatives are handled separately - see getDerivatives

    :param tdvSpec: An iterable with the names of thermodynamic variables to be evaluated
    :param dimCt:   The number of dimensions - used if setting the default tdvSpec
    :return:        An immutable iterable of TDV namedtuples that includes those specified by tdvSpec
                    and all their dependencies
    """
    # If no spec provided, use the default spec (based on # dims of the spline)
    if len(tdvSpec) == 0:
        if dimCt == 3:
            tdvSpec = list(statevarnames)  # Make mutable
        else:
            tdvSpec = [t.name for t in tdvsPTOnly]
    # TODO: try to get rid of this (horrible) next line
    tdvSpec = tdvSpec if not isinstance(tdvSpec, str) else (tdvSpec,)
    # Check for completely unsupported statevars so no one has to evaluate twice because of a typo
    unsupported = set(tdvSpec) - statevarnames
    if unsupported:
        raise ValueError('One or more unsupported statevars have been requested: ' + pformat(unsupported))

    tdvSpec = _addTDVDependencies(tdvSpec)  # Add thermodynamic variables on which requested ones depend
    return tdvSpec


#########################################
## Constants
#########################################
iP = 0; iT = 1; iM = 2  # Dimension indices
defDer = 0              # Default derivative

derivSpec = namedtuple('DerivativeSpec', ['name', 'wrtP', 'wrtT', 'wrtM'])
derivatives = _getSupportedDerivatives()
derivativenames = {d.name for d in derivatives}

statevars, statevarnames = _getSupportedThermodynamicVariables()
# [(sv.name, sv.calcFn.__name__[4:]) for sv in statevars]  # Lists TDV symbols and names

tdvsPTOnly = _getPTOnlyTDVSpec()



from math import isnan
from numbers import Number
import logging
import numpy as np
from mlbspline import load

log = logging.getLogger('lbftd')
stream = logging.StreamHandler()
stream.setFormatter(logging.Formatter('[LBFTD %(levelname)s] %(message)s'))

def loadGibbsSpline(splineFile, splineVar=None):
    """Loads a Gibbs energy spline from .mat format file
    A Gibbs energy spline should be a Matlab struct that includes the following fields:
      - sp is the main spline itself.  It must be a 2D (PT) or 3D (PTX) spline as outlined in
            mlbspline.eval.evalMultivarSpline
      - MW is a list of the molecular weights of each species in the solution, measured in kg/mol,
            such that MW[0] is the molecular weight of the solvent (even if not used, i.e., if reqMWv is false)
            and MW[1] is the molecular weight of the solute
            Field must be present and can have length of 1 (pure substance) or 2 (single solute solution).
      - nu is the number of ions in solution for the solute.  All values must be positive integers.
            If MW.size == 1 (pure substance), the value will be ignored so the field can be absent --
            otherwise the field must be present and have a numeric value.

    :param splineFile:  Full or relative path to Matlab file
    :param splineVar:   Variable to load from splineFile.
                        If not provided, the splineFile must contain exactly one variable
    :return:            A dict with the Gibbs energy spline representation required by evalGibbs functions
    """
    raw = load._stripNestingToFields(load._getRaw(splineFile, splineVar))
    sp = load.getSplineDict(load._stripNestingToValue(raw['sp']))
    load.validateSpline(sp)
    MW = _getMW(raw)
    out = {
        'sp': sp,
        'MW': MW
    }
    if MW.size > 1:  # Only add nu if dealing with a solution rather than a pure substance
        out['nu'] = _getnu(raw)
    return out

def _getMW(raw):
    try:
        MW = load._stripNestingToValue(raw['MW'])
    except KeyError:
        log.warning('Could not load MW - defaulting to empty value - will not be able to calculate some thermodynamic values')
        MW = np.empty(0)
    MW = np.array([MW]) if isinstance(MW, Number) else MW  # Wrap scalar value
    if MW.size > 2:
        raise ValueError('MW has too many elements, as multi-solute solutions are not currently supported.')
    if any([not isinstance(mw, Number) or isnan(mw) for mw in MW]):
        raise ValueError('At least one value of MW is not numeric. The entire value will be ignored.')
    return MW


def _getnu(raw):
    # Currently only supports a single solute solution so this is just an integer
    nu = load._stripNestingToValue(raw['nu'])
    if not isinstance(nu, Number) or nu != int(nu):
        raise ValueError('At least one value in nu is not an integer.')
    return int(nu)




import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='SeaFreeze',
    version='0.9.3',
    author='Marshall J. Styczinski',
    author_email='marshall.j.styczinski@jpl.nasa.gov',
    description='Thermodynamic properties of the phases of H2O',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/Bjournaux/SeaFreeze',
    packages=['seafreeze', 'mlbspline', 'lbftd'],
    package_dir={'seafreeze': 'seafreeze', 'mlbspline': 'mlbspline', 'lbftd': 'lbftd'},
    classifiers=[
        'Programming Language :: Python :: 3.8',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent'
    ],
    include_package_data=True,
    install_requires=[
        'numpy >= 1.24.2',
        'scipy >= 1.10.1',
        'psutil >= 5.9.4'
    ]
)



from scipy.interpolate import splev
import numpy as np
import logging

log = logging.getLogger('mlbspline')
stream = logging.StreamHandler()
stream.setFormatter(logging.Formatter('[MLBspline %(levelname)s] %(message)s'))

iT = 1

def evalMultivarSpline(spd, x, der=None):
    """ Performs recursive evaluation of b-spline for the given independent values
    For now, assumes 1-D spline (y for each n-D x is scalar)
    x and spd['coefs'] must have the same number of dimensions

    :param spd: a B-spline definition as given by load.loadSpline
    :param x:   Either
                    - a Numpy n-D array with the points at which the spline should be evaluated
                    or
                    - a tuple representing a single point at which the spline should be evaluated
                Either way, the number of dimensions must be same as in the spline (i.e.,
                len(x) == spd['number'].size).  Dimensions must be in the same order as in the spline.
    :param der: (optional) a list of the derivative levels for evaluation, with each value representing
                the derivative for the dimension at the corresponding index in x
                if provided, values must be non-negative integers, and
                the number of dimensions must be the same as in the spline (len(der) == spd['number'].size)
    :return:    a Numpy n-D array. If x is an ndarray, then output.shape == x.shape.
                If x is a tuple, then output.shape == tuple(np.ones(len(x), np.int)).
    """
    if der is None:
        der = []

    dimCt = spd['number'].size  # Size of dim coefs
    if len(der) == 0:
        der = [0] * dimCt  # Default to 0th derivative for all dimensions

    if len(x) != dimCt:  # Use len(x) because it works with both tuples and ndarrays
        raise ValueError('The dimensions of the evaluation points do not match the dimensions of the spline.')
    if len(der) != dimCt:
        raise ValueError('The dimensions of the derivative directives do not match the dimensions of the spline.')
    if not all((isinstance(i, int) and i >= 0) for i in der):
        raise ValueError('At least one derivative directive is not a non-negative integer.')
    # Use chain rule to get multiplicative factor for non-dimensional temperatures in B spline
    ndT1, ndT2 = (1, 1)
    if der[iT] > 0 and spd['ndT']:
        T_K = np.exp(x[iT])*spd['Tc']
        ndT1 = 1/T_K
        if der[iT] > 1:
            ndT2 = -1/T_K**2

    y = spd['coefs']
    # Start at the last index and work downward to the first
    for di in range(dimCt - 1, -1, -1):
        xi = x[di]  # Get x values for dimension being evaluated
        # Wrap xi if necessary
        if not isinstance(xi, np.ndarray):
            xi = np.asarray([xi])
        tck = _getNextSpline(di, dimCt, spd, y)
        if di == iT and spd['ndT']:
            if der[iT] == 2 and spd['ndT']:
                # Use product rule to combine derivates in terms of non-dimensional T
                y = np.array(splev(xi, tck, der=2))*ndT1**2 + np.array(splev(xi, tck, der=1))*ndT2
            else:
                y = np.array(splev(xi, tck, der=der[di]))*ndT1
        else:
            y = np.array(splev(xi, tck, der=der[di]))
    # Need to rearrange back to original order and shape
    return _getNextSpline(-1, dimCt, spd, y)[1]


def _getNextSpline(dimIdx, dimCt, spd, coefs):
    """ Get the tck (knots, coefs, and degree) for the next spline to be evaluated

    :param dimIdx:  The zero-based index of the dimension being evaluated
    :param dimCt:   The number of dimensions
    :param spd:     The spline being evaluated
    :param coefs:   The coefficients for this dimension - for this outermost dim, this is just the coefs from the spline
                    Otherwise, this is the output from the last call to interp.splev
    :return:        The [t,c,k] spline representation for the next dimension (see splev documentation)
    """
    li = dimCt - 1
    if li != dimIdx:
        coefs = np.moveaxis(coefs, li, 0)
    t = spd['knots'][dimIdx]
    k = spd['order'][dimIdx] - 1  # Scipy wants the degree, not the order (MatLab gives the orders)
    return [t, coefs, k]



from datetime import datetime
from pprint import pformat
from time import time
import logging
from collections.abc import Iterable
import numpy as np
from psutil import virtual_memory

from mlbspline.eval import evalMultivarSpline
from lbftd import statevars
from lbftd.statevars import iT, iP, iM

log = logging.getLogger('lbftd')
stream = logging.StreamHandler()
stream.setFormatter(logging.Formatter('[LBFTD %(levelname)s] %(message)s'))


def evalSolutionGibbsGrid(gibbsSp, PTM, *tdvSpec, MWv=18.01528e-3, MWu=None, failOnExtrapolate=True, verbose=False):
    """ Calculates thermodynamic variables for solutions based on a spline giving Gibbs energy
    This currently only supports single-solute solutions.

    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    Warning: units must be as specified here because some conversions are hardcoded into this function.
    With the exception of pressure, units are SI.  Pressure is in MPa rather than Pa.
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    :param gibbsSp: A B-spline  (in format given by loadSpline.loadMatSpline) for giving Gibbs energy (J/mg)
                    with dimensions pressure (MPa), temperature (K), and (optionally) molality (mol/kg), in the
                    order defined by statevars.iP, iT, and iM.
                    If molality is not provided, this function assumes that it is calculating
                    thermodynamic properties for a pure substance.
    :param PTM:     a Numpy ndarray of ndarrays with the conditions at which gibbsSp should be evaluated
                    the number of dimensions must be same as in the spline (PTM.size == gibbsSp['number'].size)
                    each of the inner ndarrays represents one of pressure (P), temperature (T), or molality (M)
                    and must be in the same order and units described in the notes for the gibbsSp parameter
                    Additionally, each dimension must be sorted from low to high values.
    :param MWv:     float with molecular weight of solvent (kg/mol).
                    Defaults to molecular weight of water (7 sig figs)
    :param MWu:     float with molecular weight of solute (kg/mol).
    :param failOnExtrapolate:   True if you want an error to appear if PTM includes values that fall outside the knot
                    sequence of gibbsSp.  If False, throws a warning rather than an error, and
                    proceeds with the calculation.
    :param verbose: boolean indicating whether to print status updates, warnings, etc.
    :param tdvSpec: iterable indicating the thermodynamic variables to be calculated
                    elements can be either strings showing the names (full list at statevars.statevarnames)
                    or TDV objects from statevars.statevars.
                    If not provided, this function will calculate the variables in statevars.tdvsPTOnly for a PT spline,
                    and those in statevars.statevars for a PTM spline.
                    Any other args provided will result in an error.
                    See the README for units.
    :return:        a named tuple with the requested thermodynamic variables as named properties
                    matching the statevars requested in the *tdvSpec parameter of this function
                    the output will also include P, T, and M (if provided) properties
    """
    [dimCt, tdvSpec] = _parseInput(gibbsSp, *tdvSpec)
    _checkInputs(gibbsSp, dimCt, tdvSpec, PTM, MWv, MWu, failOnExtrapolate)
    tdvout = createThermodynamicStatesObj(tdvSpec, PTM)  # This has the original PTM
    if _needs0M(dimCt, tdvSpec) and PTM[iM][0] != 0:
        PTM[iM] = np.insert(PTM[iM], 0, 0)
    tdvout = _evalInternal(gibbsSp, tdvSpec, PTM, MWu, MWv, tdvout, dimCt, verbose)
    return tdvout


def evalSolutionGibbsScatter(gibbsSp, PTM, *tdvSpec, MWv=18.01528e-3, MWu=None, failOnExtrapolate=True, verbose=False):
    """ Calculates thermodynamic variables for solutions based on a spline giving Gibbs energy
    This currently only supports single-solute solutions.

    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    Warning: units must be as specified here because some conversions are hardcoded into this function.
    With the exception of pressure, units are SI.  Pressure is in MPa rather than Pa.
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    :param gibbsSp: A B-spline  (in format given by loadSpline.loadMatSpline) for giving Gibbs energy (J/mg)
                    with dimensions pressure (MPa), temperature (K), and (optionally) molality (mol/kg), in the
                    order defined by statevars.iP, iT, and iM.
                    If molality is not provided, this function assumes that it is calculating
                    thermodynamic properties for a pure substance.
    :param PTM:     a Numpy ndarray of tuples with the points at which gibbsSp should be evaluated
                    the number of elements in each tuple must be same as in the spline (PTM[i].size == gibbsSp['number'].size)
                    each tuples gives the pressure (P), temperature (T), and (optionally) molality (M)
                    and must be in the same order and units described in the notes for the gibbsSp parameter
                    No sorting is required.
    :param MWv:     float with molecular weight of solvent (kg/mol).
                    Defaults to molecular weight of water (7 sig figs)
    :param MWu:     float with molecular weight of solute (kg/mol).
    :param failOnExtrapolate:   True if you want an error to appear if PTM includes values that fall outside the knot
                    sequence of gibbsSp.  If False, throws a warning rather than an error, and
                    proceeds with the calculation.
    :param verbose: boolean indicating whether to print status updates, warnings, etc.
    :param tdvSpec: iterable indicating the thermodynamic variables to be calculated
                    elements can be either strings showing the names (full list at statevars.statevarnames)
                    or TDV objects from statevars.statevars.
                    If not provided, this function will calculate the variables in statevars.tdvsPTOnly for a PT spline,
                    and those in statevars.statevars for a PTM spline.
                    Any other args provided will result in an error.
                    See the README for units.
    :return:        a named tuple with the requested thermodynamic variables as named properties
                    matching the statevars requested in the *tdvSpec parameter of this function
                    Each statevar property will contain an ndarray of values corresponding to the points listed in PTM.
                    The output also contains a PTM property for convenience.
    """
    [dimCt, tdvSpec] = _parseInput(gibbsSp, *tdvSpec)
    needs0M = _needs0M(dimCt, tdvSpec)
    fakePTM = _makeFakePTMGrid(dimCt, PTM)  # Fakes grid-style input to do all input checks before processing
    _checkInputs(gibbsSp, dimCt, tdvSpec, fakePTM, MWv, MWu, failOnExtrapolate)

    tdvout = createThermodynamicStatesObj(tdvSpec, PTM, initializetdvs=True)
    # Point by point
    for p in range(PTM.size):
        wPTM = _ptmTuple2NestedArrays(PTM[p], dimCt, needs0M)
        tempout = createThermodynamicStatesObj(tdvSpec, wPTM)
        # Prepend a 0 concentration if one is needed by any of the quantities being calculated
        if needs0M and wPTM[iM] != 0:
            wPTM[iM] = np.insert(wPTM[iM], 0, 0)
        tempout = _evalInternal(gibbsSp, tdvSpec, wPTM, MWu, MWv, tempout, dimCt, verbose=False)
        # Copy temp output to the real output
        for t in tdvSpec:
            tolastpt = getattr(tdvout, t.name)
            tolastpt[p] = getattr(tempout, t.name)
            setattr(tdvout, t.name, tolastpt)
    return tdvout


def _evalInternal(gibbsSp, tdvSpec, PTM, MWu, MWv, tdvout, dimCt, verbose=False):
    derivs = getDerivatives(gibbsSp, PTM, dimCt, tdvSpec, verbose)
    gPTM = _getGriddedPTM(tdvSpec, PTM, verbose) if any([tdv.reqGrid for tdv in tdvSpec]) else None
    f = _getVolSolventInVolSlnConversion(MWu, PTM) if any([tdv.reqF for tdv in tdvSpec]) else None

    # Calculate thermodynamic variables and store in appropriate fields in tdvout
    completedTDVs = set()  # list of completed tdvs
    while len(completedTDVs) < len(tdvSpec):
        # Get tdvs that either have empty reqTDV or all of those tdvs have already been calculated
        # but that are not in completedTDVs themselves (don't recalculate anything)
        tdvsToEval = tuple(t for t in tdvSpec if
                           t.name not in completedTDVs and (not t.reqTDV or not t.reqTDV.difference(completedTDVs)))
        for t in tdvsToEval:
            args = _buildEvalArgs(t, derivs, PTM, gPTM, MWv, MWu, tdvout, gibbsSp, f)
            start = time()
            setattr(tdvout, t.name, t.calcFn(**args))  # Calculate the value and set it in the output
            end = time()
            if verbose: _printTiming('tdv ' + t.name, start, end)
            completedTDVs.add(t.name)
    _remove0M(tdvout, PTM)
    return tdvout


def _parseInput(gibbsSp, *tdvSpec):
    dimCt = gibbsSp['number'].size
    # Expand spec to add dependencies (or set to default spec if no spec given)
    origSpec = tdvSpec
    tdvSpec = statevars.expandTDVSpec(tdvSpec, dimCt)
    addedTDVs = [s.name for s in tdvSpec if s.name not in origSpec]
    if origSpec and addedTDVs:  # The original spec was not empty and more tdvs were added
        print('NOTE: The requested thermodynamic variables depend on the following variables, which will be ' +
              'included as properties of the output object: ' + pformat(addedTDVs))
    return dimCt, tdvSpec


def _buildEvalArgs(tdv, derivs, PTM, gPTM, MWv, MWu, tdvout, gibbsSp, f):
    args = dict()
    if tdv.reqDerivs: args[tdv.parmderivs] = derivs
    if tdv.reqGrid:   args[tdv.parmgrid] = gPTM
    if tdv.reqMWv:    args[tdv.parmMWv] = MWv
    if tdv.reqMWu:    args[tdv.parmMWu] = MWu
    if tdv.reqTDV:    args[tdv.parmtdv] = tdvout
    if tdv.reqSpline: args[tdv.parmspline] = gibbsSp
    if tdv.reqPTM:    args[tdv.parmptm] = PTM
    if tdv.reqF:      args[tdv.parmf] = f
    return args


def _makeFakePTMGrid(dimCt, PTM):
    out = np.empty(dimCt, object)
    for d in np.arange(0, dimCt):
        u = {t[d] for t in PTM}  # Get unique values in each dimension
        out[d] = np.array(sorted(u))  # Sort the values in ascending order
    return out


def _checkInputs(gibbsSp, dimCt, tdvSpec, PTM, MWv, MWu, failOnExtrapolate):
    """ Checks error conditions before performing any calculations, throwing an error if anything doesn't match up
      - Ensures necessary data is available for requested statevars (check req parameters for statevars)
      - Throws a warning (or error if failOnExtrapolate=True) if spline is being evaluated on values
        outside the range of its knots
      - Warns the user if the requested output would exceed vmWarningFactor times the amount of virtual memory
      Note that the mlbspline.eval module performs additional checks (like dimension count mismatches
    """
    knotranges = [(k[0], k[-1]) for k in gibbsSp['knots']]
    # If a PTM spline, make sure the spline's concentration dimension starts with 0 if any tdv has req0M=True
    # Note that the evalGibbs functions elsewhere handle the case where PTM does not include 0 concentration.
    req0M = [t for t in tdvSpec if t.req0M]
    if dimCt == 3 and req0M and knotranges[iM][0] != 0:
        raise ValueError('You cannot calculate ' + pformat([t.name for t in req0M]) + ' with a spline that does ' +
                         'not include 0 concentration. Remove those statevars and all their dependencies, or ' +
                         'supply a spline that includes 0 concentration.')
    # Make sure that spline has 3 dims if tdvs using concentration or f are requested
    reqF = [t for t in tdvSpec if t.reqF]
    reqM = [t for t in tdvSpec if t.reqM]
    if dimCt == 2 and (reqM or reqF):
        raise ValueError('You cannot calculate ' + pformat(set([t.name for t in (reqF + reqM)])) + ' with a spline ' +
                         'that does not include concentration. Remove those statevars and all their dependencies, ' +
                         'or supply a spline that includes concentration.')
    # Make sure that solvent molecular weight is provided if any tdvs that require MWv are requested
    reqMWv = [t for t in tdvSpec if t.reqMWv]
    if (MWv == 0 or not MWv) and reqMWv:
        raise ValueError('You cannot calculate ' + pformat([t.name for t in reqMWv]) + ' without ' +
                         'providing solvent molecular weight.  Remove those statevars and all their dependencies, or ' +
                         'provide a valid value for the MWv parameter.')
    # Make sure that solute molecular weight is provided if any tdvs that require MWu or f are requested
    reqMWu = [t for t in tdvSpec if t.reqMWu]
    if (MWu == 0 or not MWu) and (reqMWu or reqF):
        raise ValueError('You cannot calculate ' + pformat([t.name for t in set(reqMWu + reqF)]) + ' without ' +
                         'providing solute molecular weight.  Remove those statevars and all their dependencies, or ' +
                         'provide a valid value for the MWu parameter.')
    # Make sure that all the PTM values fall inside the knot ranges of the spline
    # For single point, PTM is a tuple - just report its min and max as the single value
    # For a grid, PTM is an ndarray of sorted ndarrays, and so the min and max are the first and last elements
    # TODO: make this less explictly dependent on data types
    ptmranges = [(d[0], d[-1]) if isinstance(d, Iterable) else (d, d) for d in PTM]
    hasValsOutsideKnotRange = lambda kr, dr: dr[0] < kr[0] or dr[1] > kr[1]
    extrapolationDims = [i for i in range(0, dimCt) if hasValsOutsideKnotRange(knotranges[i], ptmranges[i])]
    if extrapolationDims:
        msg = ' '.join(
            ['Dimensions', pformat({'P' if d == iP else ('T' if d == iT else 'M') for d in extrapolationDims}),
             'contain values that fall outside the knot sequence for the given spline,',
             'which will result in extrapolation, which may not produce meaningful values.'])
        if failOnExtrapolate:
            raise ValueError(msg)
        else:
            log.warning(msg)
    # Warn the user if the calculation results will take more than some factor times total virtual memory
    # TODO: make this less explicitly dependent on data types
    ptct = np.prod([len(d) if isinstance(d, Iterable) else 1 for d in PTM])  # the total number of points
    # For each point, the output will include 1 value for the PTM point itself plus 1 for each tdv
    outputSize = (len(tdvSpec) + 1) * ptct * floatSizeBytes
    if outputSize > virtual_memory().total * vmWarningFactor:
        log.warning(f'The projected output is more than {vmWarningFactor} times the total virtual memory for this machine.')
    return


def createThermodynamicStatesObj(tdvSpec, PTM, initializetdvs=False):
    """

    :param tdvSpec:         The tdvSpec indicating which tdvs will need to be stored in the object
    :param PTM:             The original PTM data, which will be stored in the object
    :param initializetdvs:  True if creating the ThermodynamicStates object for scatter data
                            Initializes output in the object by creating empty arrays matching the size of PTM
    :return:                A ThermodynamicStates object ready to store calculated thermodynamic state variables
    """
    flds = {t.name for t in tdvSpec} | {'PTM'}

    # Copy PTM so if you add a 0 concentration, you affect only the version in the output var
    # so later you can compare the original PTM to the one in tdvout.PTM to see if you need to remove the 0 M
    TDS = type('ThermodynamicStates', (object,),
               {fld: (np.copy(PTM) if fld == 'PTM' else np.empty((PTM.size,), float) if initializetdvs else None) for
                fld in flds})
    out = TDS()
    return out


def _ptmTuple2NestedArrays(PTM, dimCt, needs0M):
    """ Converts a PTM tuple to numpy ndarrays

    :return: PTM as nested arrays, possibly with introduced 0 M
    """
    out = np.empty(dimCt, object)
    for i in np.arange(0, dimCt):
        out[i] = np.empty((1,), float)
        out[i][0] = PTM[i]
    return out


def _needs0M(dimCt, tdvSpec):
    """ Determines whether any tdvs require 0 concentration to be properly evaluated
    """
    return dimCt == 3 and any([t.req0M for t in tdvSpec])


def _remove0M(tdvout, wrappedPTM):
    """ If a 0 concentration was added, take it back out.
    NOTE: This method changes the value of tdvout without returning it.

    :param tdvout:      The object with calculated thermodynamic properties
                        tdvout.PTM is the original PTM input provided by the caller
    :param wrappedPTM:  The wrapped PTM which may include
    """
    if wrappedPTM.size == 3 and wrappedPTM[iM][0] == 0:
        try:  # tdvout.PTM is either a tuple or another set of nested arrays
            firstM = tdvout.PTM[iM][0]
        except IndexError:
            firstM = tdvout.PTM[iM]
        if firstM != 0:  # This means the wrapped PTM had a 0 concentration prepended
            # Go through all calculated values and remove the first item from the M dimension
            # TODO: figure out why PTM doesn't show up in vars
            for p, v in vars(tdvout).items():
                slc = [slice(None)] * len(v.shape)
                slc[iM] = slice(1, None)
                setattr(tdvout, p, v[tuple(slc)])


def _createGibbsDerivativesClass(tdvSpec):
    flds = {d for t in tdvSpec if t.reqDerivs for d in t.reqDerivs}
    return type('GibbsDerivatives', (object,), {d: None for t in tdvSpec if t.reqDerivs for d in t.reqDerivs})


def _buildDerivDirective(derivSpec, dimCt):
    """ Gets a list of the derivatives for relevant dimensions

    :param derivSpec: Derivatives specified
    :param dimCt: int. Dimension count (2 or 3)
    :return:
        out. Derivative specification
    """
    out = [statevars.defDer] * dimCt
    if derivSpec.wrtP: out[iP] = derivSpec.wrtP
    if derivSpec.wrtT: out[iT] = derivSpec.wrtT
    if derivSpec.wrtM: out[iM] = derivSpec.wrtM
    return out


def getDerivatives(gibbsSp, PTM, dimCt, tdvSpec, verbose=False):
    """

    :param gibbsSp:     The Gibbs energy spline
    :param PTM:         The pressure, temperature[, molality] points at which the spine should be evaluated
    :param dimCt:       2 if a PT spline, 3 if a PTM spline
    :param tdvSpec:     The expanded TDVSpec describing the thermodynamic variables to be calculated
    :param verbose:     True to output timings
    :return:            Gibbs energy derivatives necessary to calculate the tdvs listed in the tdvSpec
    """
    GibbsDerivs = _createGibbsDerivativesClass(tdvSpec)
    out = GibbsDerivs()
    reqderivs = {d for t in tdvSpec for d in t.reqDerivs}  # Get set of derivative names that are needed
    getDerivSpec = lambda dn: next(d for d in statevars.derivatives if d.name == dn)
    for rd in reqderivs:
        derivDirective = _buildDerivDirective(getDerivSpec(rd), dimCt)
        start = time()
        setattr(out, rd, evalMultivarSpline(gibbsSp, PTM, derivDirective))
        end = time()
        if verbose: _printTiming('deriv ' + rd, start, end)
    return out


def _getGriddedPTM(tdvSpec, PTM, verbose=False):
    if any([t.reqGrid for t in tdvSpec]):
        start = time()
        out = np.meshgrid(*PTM.tolist(), indexing='ij')  # Grid the dimensions of PTM
        end = time()
        if verbose: _printTiming('grid', start, end)
    else:
        out = []
    return out


def _getVolSolventInVolSlnConversion(MWu, PTM):
    """ Dimensionless conversion factor for how much of the volume of 1 kg of solution is really just the solvent
    :return: f
    """
    return 1 + MWu * PTM[iM]


def _printTiming(calcdesc, start, end):
    endDT = datetime.fromtimestamp(end)
    print(endDT.strftime('%H:%M:%S.%f'), ':\t', calcdesc, 'took', str(end - start), 'seconds to calculate')


#########################################
## Constants
#########################################
vmWarningFactor = 2  # Warn the user when size of output would exceed vmWarningFactor times total virtual memory
floatSizeBytes = int(np.finfo(float).bits / 8)
